@@ -1,9 +1,12 @@
+import datetime
+import warnings
+
 import pytest
 
 import pg8000
 
 
-def testUnixSocketMissing():
+def test_unix_socket_missing():
     conn_params = {"unix_sock": "/file-does-not-exist", "user": "doesn't-matter"}
 
     with pytest.raises(pg8000.dbapi.InterfaceError):
@@ -21,12 +24,74 @@ def test_internet_socket_connection_refused():
         pg8000.dbapi.connect(**conn_params)
 
 
-def testDatabaseMissing(db_kwargs):
+def test_database_missing(db_kwargs):
     db_kwargs["database"] = "missing-db"
     with pytest.raises(pg8000.dbapi.DatabaseError):
         pg8000.dbapi.connect(**db_kwargs)
 
 
+def test_database_name_unicode(db_kwargs):
+    db_kwargs["database"] = "pg8000_sn\uFF6Fw"
+
+    # Should only raise an exception saying db doesn't exist
+    with pytest.raises(pg8000.dbapi.DatabaseError, match="3D000"):
+        pg8000.dbapi.connect(**db_kwargs)
+
+
+def test_database_name_bytes(db_kwargs):
+    """Should only raise an exception saying db doesn't exist"""
+
+    db_kwargs["database"] = bytes("pg8000_sn\uFF6Fw", "utf8")
+    with pytest.raises(pg8000.dbapi.DatabaseError, match="3D000"):
+        pg8000.dbapi.connect(**db_kwargs)
+
+
+def test_password_bytes(con, db_kwargs):
+    # Create user
+    username = "boltzmann"
+    password = "cha\uFF6Fs"
+    cur = con.cursor()
+    cur.execute("create user " + username + " with password '" + password + "';")
+    con.commit()
+
+    db_kwargs["user"] = username
+    db_kwargs["password"] = password.encode("utf8")
+    db_kwargs["database"] = "pg8000_md5"
+    with pytest.raises(pg8000.dbapi.DatabaseError, match="3D000"):
+        pg8000.dbapi.connect(**db_kwargs)
+
+    cur.execute("drop role " + username)
+    con.commit()
+
+
+def test_application_name(db_kwargs):
+    app_name = "my test application name"
+    db_kwargs["application_name"] = app_name
+    with pg8000.dbapi.connect(**db_kwargs) as db:
+        cur = db.cursor()
+        cur.execute(
+            "select application_name from pg_stat_activity "
+            " where pid = pg_backend_pid()"
+        )
+
+        application_name = cur.fetchone()[0]
+        assert application_name == app_name
+
+
+def test_application_name_integer(db_kwargs):
+    db_kwargs["application_name"] = 1
+    with pytest.raises(
+        pg8000.dbapi.InterfaceError,
+        match="The parameter application_name can't be of type " "<class 'int'>.",
+    ):
+        pg8000.dbapi.connect(**db_kwargs)
+
+
+def test_application_name_bytearray(db_kwargs):
+    db_kwargs["application_name"] = bytearray(b"Philby")
+    pg8000.dbapi.connect(**db_kwargs)
+
+
 def test_notify(con):
     cursor = con.cursor()
     cursor.execute("select pg_backend_pid()")
@@ -55,40 +120,6 @@ def test_notify_with_payload(con):
     assert con.notifications[0] == (backend_pid, "test", "Parnham")
 
 
-def testUnicodeDatabaseName(db_kwargs):
-    db_kwargs["database"] = "pg8000_sn\uFF6Fw"
-
-    # Should only raise an exception saying db doesn't exist
-    with pytest.raises(pg8000.dbapi.DatabaseError, match="3D000"):
-        pg8000.dbapi.connect(**db_kwargs)
-
-
-def testBytesDatabaseName(db_kwargs):
-    """Should only raise an exception saying db doesn't exist"""
-
-    db_kwargs["database"] = bytes("pg8000_sn\uFF6Fw", "utf8")
-    with pytest.raises(pg8000.dbapi.DatabaseError, match="3D000"):
-        pg8000.dbapi.connect(**db_kwargs)
-
-
-def testBytesPassword(con, db_kwargs):
-    # Create user
-    username = "boltzmann"
-    password = "cha\uFF6Fs"
-    cur = con.cursor()
-    cur.execute("create user " + username + " with password '" + password + "';")
-    con.commit()
-
-    db_kwargs["user"] = username
-    db_kwargs["password"] = password.encode("utf8")
-    db_kwargs["database"] = "pg8000_md5"
-    with pytest.raises(pg8000.dbapi.DatabaseError, match="3D000"):
-        pg8000.dbapi.connect(**db_kwargs)
-
-    cur.execute("drop role " + username)
-    con.commit()
-
-
 def test_broken_pipe_read(con, db_kwargs):
     db1 = pg8000.dbapi.connect(**db_kwargs)
     cur1 = db1.cursor()
@@ -130,29 +161,46 @@ def test_broken_pipe_unpack(con):
         cur.execute("select pg_terminate_backend(%s)", (pid1,))
 
 
-def testApplicatioName(db_kwargs):
-    app_name = "my test application name"
-    db_kwargs["application_name"] = app_name
-    with pg8000.dbapi.connect(**db_kwargs) as db:
-        cur = db.cursor()
-        cur.execute(
-            "select application_name from pg_stat_activity "
-            " where pid = pg_backend_pid()"
-        )
+def test_py_value_fail(con, mocker):
+    # Ensure that if types.py_value throws an exception, the original
+    # exception is raised (PG8000TestException), and the connection is
+    # still usable after the error.
 
-        application_name = cur.fetchone()[0]
-        assert application_name == app_name
+    class PG8000TestException(Exception):
+        pass
+
+    def raise_exception(val):
+        raise PG8000TestException("oh noes!")
+
+    mocker.patch.object(con, "py_types")
+    con.py_types = {datetime.time: (1083, raise_exception)}
+
+    with pytest.raises(PG8000TestException):
+        c = con.cursor()
+        c.execute("SELECT %s as f1", (datetime.time(10, 30),))
+        c.fetchall()
+
+        # ensure that the connection is still usable for a new query
+        c.execute("VALUES ('hw3'::text)")
+        assert c.fetchone()[0] == "hw3"
 
 
-def test_application_name_integer(db_kwargs):
-    db_kwargs["application_name"] = 1
-    with pytest.raises(
-        pg8000.dbapi.InterfaceError,
-        match="The parameter application_name can't be of type " "<class 'int'>.",
-    ):
-        pg8000.dbapi.connect(**db_kwargs)
+def test_no_data_error_recovery(con):
+    for i in range(1, 4):
+        with pytest.raises(pg8000.DatabaseError) as e:
+            c = con.cursor()
+            c.execute("DROP TABLE t1")
+        assert e.value.args[0]["C"] == "42P01"
+        con.rollback()
 
 
-def test_application_name_bytearray(db_kwargs):
-    db_kwargs["application_name"] = bytearray(b"Philby")
-    pg8000.dbapi.connect(**db_kwargs)
+def test_closed_connection(db_kwargs):
+    warnings.simplefilter("ignore")
+
+    my_db = pg8000.connect(**db_kwargs)
+    cursor = my_db.cursor()
+    my_db.close()
+    with pytest.raises(my_db.InterfaceError, match="connection is closed"):
+        cursor.execute("VALUES ('hw1'::text)")
+
+    warnings.resetwarnings()
